import time
import threading
import logging
from .config import config
from .core import REPEAT_QUEUE


class LogParser(object):

    request_url = None
    request_method = "get"
    request_headers = None
    request_body = None
    request_start_timestamp = None  # unix时间戳,毫秒级
    is_matched = False

    def __init__(self, content):
        self.content = content

    def parse(self):
        """
        子类须实现该方法,如果记录匹配,必须修改is_matched为True
        :return:
        """
        pass

    def obj_to_dict(self):
        if self.is_matched:
            return dict(
                url=self.request_url, method=self.request_method, body=self.request_body, headers=self.request_headers)
        return {}


class ParserThread(threading.Thread):

    def __init__(self, log_file, log_parser, file_encoding='utf-8'):
        super(ParserThread, self).__init__()
        if not issubclass(log_parser, LogParser):
            raise TypeError
        self.log_file = log_file
        self.log_parser = log_parser
        self.out_q = REPEAT_QUEUE
        self.logger = logging.getLogger(__name__)
        self.file_encoding = file_encoding

    def run(self):
        last_gather_ts = None
        diff_ts = None  # 回放时间与原记录时间差
        matched_records_in_cycle = 0

        with open(self.log_file, "r", encoding=self.file_encoding) as f:

            for line in f.readlines():
                parsed = self.log_parser(line)
                parsed.parse()
                if not parsed.is_matched:  # 当该行并非请求日志时,不处理
                    continue

                if config.GATHER_INTERVAL == 0:  # 不控制采集间隔的情况下, 不会控制回放的节奏
                    self.logger.debug(parsed.obj_to_dict())
                    self.out_q.async_q.put_nowait(parsed.obj_to_dict())
                else:
                    if parsed.request_start_timestamp is None:
                        raise ValueError("if set GATHER_INTERVAL not eq 0, must specify the `request_start_timestamp` "
                                         "in `parse()`")
                    if last_gather_ts is None:  # 取到第一条匹配的请求日志, 回放开始
                        last_gather_ts = parsed.request_start_timestamp
                        diff_ts = int(time.time() * 1000) - last_gather_ts
                        matched_records_in_cycle += 1
                        self.logger.debug(parsed.obj_to_dict())
                        self.out_q.async_q.put_nowait(parsed.obj_to_dict())
                    else:
                        if parsed.request_start_timestamp - last_gather_ts <= config.GATHER_INTERVAL * 1000:
                            self.logger.debug(parsed.obj_to_dict())
                            matched_records_in_cycle += 1
                            self.out_q.async_q.put_nowait(parsed.obj_to_dict())
                        else:
                            while 1:
                                self.logger.info("gathered {} records in {} seconds".format(
                                    matched_records_in_cycle, config.GATHER_INTERVAL))
                                time.sleep(config.GATHER_INTERVAL)
                                last_gather_ts = int(time.time()*1000) - diff_ts
                                matched_records_in_cycle = 0

                                if parsed.request_start_timestamp - last_gather_ts < config.GATHER_INTERVAL * 1000:
                                    matched_records_in_cycle += 1
                                    self.logger.debug(parsed.obj_to_dict())
                                    self.out_q.async_q.put_nowait(parsed.obj_to_dict())
                                    break

        self.logger.info("read complete")




import asyncio
import aiohttp
import random
import logging
import janus
from .config import config
from concurrent.futures import ThreadPoolExecutor
import timeit


EVENT_LOOP = asyncio.get_event_loop()
REPLAY_QUEUE = janus.Queue(loop=EVENT_LOOP)  # 回放队列,生产者为中继器,消费者将会生成的异步请求
REPEAT_QUEUE = janus.Queue(loop=EVENT_LOOP)  # 中继队列,消费者为中继器,用于更改请求量
EXECUTOR = ThreadPoolExecutor(config.THREAD_POOL_NUMBER)
CLIENT = aiohttp.ClientSession(loop=EVENT_LOOP)
LOGGER = logging.getLogger(__name__)


async def repeater(repeat_q, replay_q, rate):
    """
    中继器
    :param repeat_q:
    :param replay_q:
    :param rate:
    :return:
    """
    while 1:
        parameters = await repeat_q.async_q.get()
        loop = rate
        while loop > 0:
            if loop >= 1:
                replay_q.async_q.put_nowait(parameters.copy())
            else:
                r = random.random()
                if r <= rate:
                    replay_q.async_q.put_nowait(parameters.copy())
            loop -= 1


async def request(client, method, url, **kwargs):
    """
    http请求函数
    :param method:
    :param url:
    :param kwargs:
    :return:
    """
    if config.RESPONSE_HANDLER is not None:
        future = asyncio.Future()
        future.add_done_callback(config.RESPONSE_HANDLER)
    start = timeit.default_timer()
    async with client.request(method, url, **kwargs) as response:
        content = await response.text()
        r = dict(
            request=dict(url=url, method=method, paramenters=kwargs),
            response=dict(
                status_code=response.status,
                content=content,
                elapsed=timeit.default_timer()-start,
                host=response.host,
                method=response.method,
                url=response.url,
                cokkies=response.cookies
            ))
        if config.RESPONSE_HANDLER is not None:
            future.set_result(r)


async def player(q):
    """
    请求发送器
    :param q:
    :return:
    """
    while 1:
        parameters = await q.async_q.get()
        method = parameters.pop('method', 'get')
        url = parameters.pop('url')
        params = parameters.pop('param', None)
        data = parameters.pop('body', None)
        headers = parameters.pop('headers', None)
        await request(CLIENT, method, url, params=params, data=data, headers=headers, **parameters)



import os
import asyncio
from .config import config
from .core import EVENT_LOOP, REPEAT_QUEUE, REPLAY_QUEUE, player, repeater, EXECUTOR
from .parser import ParserThread
from .monitor import MonitorThread


def main(log_file, log_parser, rate=1, file_encoding="utf-8", callback=None):
    """
    logreplay主函数
    :param log_file: 日志文件
    :param rate: 默认原速回; 当>1时, 比如2,从日志中读取到的一条http请求,将会被并发请求两次; 当0<rate<1时,如0.5,该http请求,有50%的概率不回放
    :param log_parser: 日志解析器
    :param file_encoding:
    :return:
    """

    if not os.path.isfile(log_file):
        raise IOError("cannot find log file: {}".format(log_file))

    if callback is not None:
        config.RESPONSE_HANDLER = callback

    pt = ParserThread(log_file, log_parser, file_encoding=file_encoding)
    pt.start()
    MonitorThread(pt).start()

    [asyncio.ensure_future(repeater(REPEAT_QUEUE, REPLAY_QUEUE, rate)) for _ in range(config.REPEATER_NUMBER)]
    [asyncio.ensure_future(player(REPLAY_QUEUE)) for _ in range(config.PLAYER_NUMBER)]

    EVENT_LOOP.set_default_executor(EXECUTOR)
    EVENT_LOOP.run_forever()



import logging
import threading
import time
from .core import EVENT_LOOP, REPLAY_QUEUE, REPEAT_QUEUE, EXECUTOR, CLIENT


class MonitorThread(threading.Thread):

    def __init__(self, monitored):
        super(MonitorThread, self).__init__()
        self.monitored = monitored
        self.logger = logging

    def run(self):
        while self.monitored.isAlive():
            time.sleep(.5)
        self.logger.info("the monitored thread is dead")

        # close REPEAT_QUEUE
        while REPEAT_QUEUE.async_q.qsize():
            time.sleep(.5)
        else:
            REPEAT_QUEUE.close()
        self.logger.info("closed REPEAT_QUEUE")

        # close REPLAY_QUEUE
        while REPLAY_QUEUE.async_q.qsize():
            time.sleep(.5)
        else:
            REPLAY_QUEUE.close()
        self.logger.info("closed REPLAY_QUEUE")

        # close loop
        EXECUTOR.shutdown(wait=True)
        self.logger.info('shutdown executor')
        CLIENT.close()
        EVENT_LOOP.stop()
        try:
            EVENT_LOOP.close()
        except RuntimeError:
            self.logger.warning("close event loop fail, press CTRL+C to shutdown log-replay")
